﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Musiclibrary {
    /// <summary>
    /// Provides functions to share the local library in a LAN.
    /// </summary>
    public class SharingServer {
        protected Thread ReplyToBroadcastsThread, AcceptThread;

        public SharingServer() {
            this.Clients = new List<ClientConnection>();
        }

        /// <summary>
        /// Starts sharing the local library.
        /// </summary>
        public void Start() {
            if (!this.IsRunning) {
                this.ReplyToBroadcastsThread = new Thread(new ThreadStart(this.ReplyToBroadcasts));
                this.AcceptThread = new Thread(new ThreadStart(this.AcceptConnections));

                this.ReplyToBroadcastsThread.Start();
                this.AcceptThread.Start();

                this.IsRunning = true;
            }
        }

        /// <summary>
        /// Disconnects all clients and stops sharing.
        /// </summary>
        public void Stop() {
            foreach (var client in this.Clients)
                client.Disconnect();

            if (this.IsRunning) {
                this.ReplyToBroadcastsThread.Abort();
                this.AcceptThread.Abort();

                try {
                    var dummySocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    dummySocket.Connect(new IPEndPoint(IPAddress.Loopback, 16790));
                    dummySocket.Close();
                } catch { }

                this.IsRunning = false;
            }
        }

        public bool IsRunning { get; protected set; }

        public List<ClientConnection> Clients { get; protected set; }

        /// <summary>
        /// This function replies to all broadcasts looking for other music libraries.
        /// </summary>
        protected void ReplyToBroadcasts() {
            //Initialize socket at port 16789
            var serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            var localEndPoint = new IPEndPoint(IPAddress.Any, 16789);


            try {
                serverSocket.ReceiveTimeout = 1000;
                serverSocket.Bind(localEndPoint);

                while (true)
                    try {
                        //Receive message
                        byte[] buffer = new byte[1024];
                        EndPoint remoteEndPoint = new IPEndPoint(IPAddress.Loopback, 0);
                        serverSocket.ReceiveFrom(buffer, ref remoteEndPoint);
                        var message = Encoding.UTF8.GetString(buffer).Trim((char)0);

                        //Is music library discovery message?
                        if (message.StartsWith("Hello! Any music library there?")) {
                            //Create new socket for replying
                            var clientSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);

                            //Reply
                            var reply = "Yoo! I am " + App.This.Library.Name;
                            buffer = UTF8Encoding.UTF8.GetBytes(reply);
                            clientSocket.SendTo(buffer, remoteEndPoint);

                            //Close connection
                            clientSocket.Close();
                        }
                    } catch (ThreadAbortException) {
                        throw; //If thread should be aborted, go to the big try..catch
                    }
                    catch { } //Ignore eventual socket errors
            } 
            catch { } //Catch socket binding errors
            finally {
                if (serverSocket.IsBound)
                    serverSocket.Close();
            }
        }

        protected Socket ServerSocket;

        /// <summary>
        /// Accepts incoming client connections.
        /// </summary>
        protected void AcceptConnections() {
            //Listen at port 16790
            this.ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            var localEndPoint = new IPEndPoint(IPAddress.Any, 16790);

            try {
                this.ServerSocket.Bind(localEndPoint);
                
                while (true)
                    try {
                        //Accept connection
                        this.ServerSocket.Listen(1);
                        var clientSocket = this.ServerSocket.Accept();

                        //Raise event
                        this.OnClientConnected(clientSocket);
                    } catch (ThreadAbortException) {
                        throw;
                    } catch { }
            } 
            catch { } 
            finally {
                if (this.ServerSocket.IsBound)
                    this.ServerSocket.Close();
            }
        }

        protected void OnClientConnected(Socket socket) {
            //Create client connection (which automatically replies to commands)
            //and add client to list
            this.Clients.Add(new ClientConnection(socket));
        }
    }
}
