﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WMPLib;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Proxies;
using System.IO;
using System.Windows.Input;

namespace Musiclibrary {
    /// <summary>
    /// An implementation of IPlayer using the Windows Media Player API
    /// </summary>
    // See IPlayer interface for some more documentation
    public class WMPPlayer : IPlayer {
        protected WindowsMediaPlayer Player = new WindowsMediaPlayerClass();

        public WMPPlayer() {
            this.Playlist = new List<ISong>();
            this.Player.PlayStateChange += new _WMPOCXEvents_PlayStateChangeEventHandler(e =>
                this.OnStateChanged());
        }

        #region IPlayer Member

        public List<ISong> Playlist {
            get;
            private set;
        }

        private int _songIndex;
        public int SongIndex {
            get { return this._songIndex; }
            set {
                this._songIndex = value;

                if (this.Playlist.Count > value)
                    this.Song = this.Playlist[value];
                else
                    this.Song = null;
            }
        }

        private ISong _song;
        public ISong Song {
            get {
                return this._song;
            }
            set {
                this._song = value;

                if (value != null) {
                    Mouse.OverrideCursor = Cursors.Wait;
                    this.Player.URL = value.Load();
                    Mouse.OverrideCursor = null;
                } else
                    this.Player.controls.pause();

                this.OnSongChanged();
                this.OnSongIndexChanged();
            }
        }

        public event Action<IPlayer, EventArgs> SongChanged;

        protected void OnSongChanged() {
            if (this.SongChanged != null)
                this.SongChanged(this, EventArgs.Empty);
        }

        public event Action<IPlayer, EventArgs> SongIndexChanged;

        protected void OnSongIndexChanged() {
            if (this.SongIndexChanged != null)
                this.SongIndexChanged(this, EventArgs.Empty);
        }

        public PlayerState State {
            get {
                if (this.Player.playState == WMPPlayState.wmppsPlaying || this._songChange)
                    return PlayerState.Playing;
                else if (this.Player.playState == WMPPlayState.wmppsPaused)
                    return PlayerState.Paused;
                else
                    return PlayerState.Finished;
            }
        }

        public event Action<IPlayer, EventArgs> StateChanged;

        protected void OnStateChanged() {
            if (this.Player.playState == WMPPlayState.wmppsMediaEnded) {
                //If song is finished, go to next song.. but wait a short time before! This is because... I can't remember.
                new TimeSpan(0, 0, 0, 0, 100).After(() => this.Next(), MainWindow.Instance);
            }
           
            else if (this.StateChanged != null && !this._songChange)
                this.StateChanged(this, EventArgs.Empty);
        }

        public TimeSpan Position {
            get {
                var position = this.Player.controls.currentPosition;
                var timespan = new TimeSpan(0, 0, 0, (int)position, (int)((position - Math.Floor(position)) * 1000));
                return timespan;
            }
            set {
                this.Player.controls.currentPosition = value.TotalSeconds;
            }
        }

        public TimeSpan Duration {
            get {
                var duration = this.Player.currentMedia.duration;
                var timespan = new TimeSpan(0, 0, 0, (int)duration, (int)((duration - Math.Floor(duration)) * 1000));
                return timespan;
            }
        }


        public double Volume {
            get {
                return this.Player.settings.volume / 100d;
            }
            set {
                this.Player.settings.volume = (int)(value * 100);
            }
        }

        public void Play() {
            this.Player.controls.play();
        }

        public void Pause() {
            this.Player.controls.pause();
        }

        private bool _songChange;

        public void Next() {
            _songChange = true;

            if (this.SongIndex < this.Playlist.Count - 1)
                this.SongIndex++;
            else if (this.Repeat) {
                this.SongIndex = 0;
            }

            _songChange = false;
        }

        public void Previous() {
            _songChange = true;

            if (this.SongIndex > 0)
                this.SongIndex--;
            else if (this.Repeat)
                this.SongIndex = this.Playlist.Count - 1;

            _songChange = false;
        }

        private bool _repeat;
        public bool Repeat {
            get { return this._repeat; }
            set {
                this._repeat = value;
                this.OnSongIndexChanged();
            }
        }

        #endregion

        #region IPlayer Member

        public void Shuffle() {
            var gen = new Random();
            var songs = from i in this.Playlist.ToList() orderby gen.Next() select i;

            this.Playlist.Clear();
            this.Playlist.AddRange(songs);

            this._songIndex = this.Playlist.IndexOf(this.Song);
            this.OnSongIndexChanged();
        }

        #endregion
    }
}
